import time
import logging
import numpy as np
import emcee

import torch
import torch.nn as nn
import torch.optim as optim

from scipy import optimize

from pybnn.base_model import BaseModel
from pybnn.util.normalization import zero_mean_unit_var_normalization, zero_mean_unit_var_denormalization
from pybnn.bayesian_linear_regression import BayesianLinearRegression, Prior


class Net(nn.Module):
    def __init__(self, n_inputs, n_units=[50, 50, 50]):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(n_inputs, n_units[0])
        self.fc2 = nn.Linear(n_units[0], n_units[1])
        self.fc3 = nn.Linear(n_units[1], n_units[2])
        self.out = nn.Linear(n_units[2], 1)

    def forward(self, x):
        x = torch.tanh(self.fc1(x))
        x = torch.tanh(self.fc2(x))
        x = torch.tanh(self.fc3(x))

        return self.out(x)

    def basis_funcs(self, x):
        x = torch.tanh(self.fc1(x))
        x = torch.tanh(self.fc2(x))
        x = torch.tanh(self.fc3(x))
        return x


class DNGO(BaseModel):

    def __init__(self, batch_size=10, num_epochs=500,
                 learning_rate=0.01,
                 adapt_epoch=5000, n_units_1=50, n_units_2=50, n_units_3=50,
                 alpha=1.0, beta=1000, prior=None, do_mcmc=True,
                 n_hypers=20, chain_length=2000, burnin_steps=2000,
                 normalize_input=True, normalize_output=True, rng=None):
        """
        Deep Networks for Global Optimization [1]. This module performs
        Bayesian Linear Regression with basis function extracted from a
        feed forward neural network.

        [1] J. Snoek, O. Rippel, K. Swersky, R. Kiros, N. Satish,
            N. Sundaram, M.~M.~A. Patwary, Prabhat, R.~P. Adams
            Scalable Bayesian Optimization Using Deep Neural Networks
            Proc. of ICML'15

        Parameters
        ----------
        batch_size: int
            Batch size for training the neural network
        num_epochs: int
            Number of epochs for training
        learning_rate: float
            Initial learning rate for Adam
        adapt_epoch: int
            Defines after how many epochs the learning rate will be decayed by a factor 10
        n_units_1: int
            Number of units in layer 1
        n_units_2: int
            Number of units in layer 2
        n_units_3: int
            Number of units in layer 3
        alpha: float
            Hyperparameter of the Bayesian linear regression
        beta: float
            Hyperparameter of the Bayesian linear regression
        prior: Prior object
            Prior for alpa and beta. If set to None the default prior is used
        do_mcmc: bool
            If set to true different values for alpha and beta are sampled via MCMC from the marginal log likelihood
            Otherwise the marginal log likehood is optimized with scipy fmin function
        n_hypers : int
            Number of samples for alpha and beta
        chain_length : int
            The chain length of the MCMC sampler
        burnin_steps: int
            The number of burnin steps before the sampling procedure starts
        normalize_output : bool
            Zero mean unit variance normalization of the output values
        normalize_input : bool
            Zero mean unit variance normalization of the input values
        rng: np.random.RandomState
            Random number generator
        """

        if rng is None:
            self.rng = np.random.RandomState(np.random.randint(0, 10000))
        else:
            self.rng = rng

        self.X = None
        self.y = None
        self.network = None
        self.alpha = alpha
        self.beta = beta
        self.normalize_input = normalize_input
        self.normalize_output = normalize_output

        # MCMC hyperparameters
        self.do_mcmc = do_mcmc
        self.n_hypers = n_hypers
        self.chain_length = chain_length
        self.burned = False
        self.burnin_steps = burnin_steps
        if prior is None:
            self.prior = Prior(rng=self.rng)
        else:
            self.prior = prior

        # Network hyper parameters
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.init_learning_rate = learning_rate

        self.n_units_1 = n_units_1
        self.n_units_2 = n_units_2
        self.n_units_3 = n_units_3
        self.adapt_epoch = adapt_epoch
        self.network = None
        self.models = []
        self.hypers = None

    @BaseModel._check_shapes_train
    def train(self, X, y, do_optimize=True):
        """
        Trains the model on the provided data.

        Parameters
        ----------
        X: np.ndarray (N, D)
            Input data points. The dimensionality of X is (N, D),
            with N as the number of points and D is the number of features.
        y: np.ndarray (N,)
            The corresponding target values.
        do_optimize: boolean
            If set to true the hyperparameters are optimized otherwise
            the default hyperparameters are used.

        """
        start_time = time.time()

        # Normalize inputs
        if self.normalize_input:
            self.X, self.X_mean, self.X_std = zero_mean_unit_var_normalization(X)
        else:
            self.X = X

        # Normalize ouputs
        if self.normalize_output:
            self.y, self.y_mean, self.y_std = zero_mean_unit_var_normalization(y)
        else:
            self.y = y

        self.y = self.y[:, None]

        # Check if we have enough points to create a minibatch otherwise use all data points
        if self.X.shape[0] <= self.batch_size:
            batch_size = self.X.shape[0]
        else:
            batch_size = self.batch_size

        # Create the neural network
        features = X.shape[1]

        self.network = Net(n_inputs=features, n_units=[self.n_units_1, self.n_units_2, self.n_units_3])

        optimizer = optim.Adam(self.network.parameters(),
                               lr=self.init_learning_rate)

        # Start training
        lc = np.zeros([self.num_epochs])
        for epoch in range(self.num_epochs):

            epoch_start_time = time.time()

            train_err = 0
            train_batches = 0

            for batch in self.iterate_minibatches(self.X, self.y,
                                                  batch_size, shuffle=True):
                inputs = torch.Tensor(batch[0])
                targets = torch.Tensor(batch[1])

                optimizer.zero_grad()
                output = self.network(inputs)
                loss = torch.nn.functional.mse_loss(output, targets)
                loss.backward()
                optimizer.step()

                train_err += loss
                train_batches += 1

            lc[epoch] = train_err / train_batches
            logging.debug("Epoch {} of {}".format(epoch + 1, self.num_epochs))
            curtime = time.time()
            epoch_time = curtime - epoch_start_time
            total_time = curtime - start_time
            logging.debug("Epoch time {:.3f}s, total time {:.3f}s".format(epoch_time, total_time))
            logging.debug("Training loss:\t\t{:.5g}".format(train_err / train_batches))

        # Design matrix
        self.Theta = self.network.basis_funcs(torch.Tensor(self.X)).data.numpy()

        if do_optimize:
            if self.do_mcmc:
                self.sampler = emcee.EnsembleSampler(self.n_hypers, 2,
                                                     self.marginal_log_likelihood)

                # Do a burn-in in the first iteration
                if not self.burned:
                    # Initialize the walkers by sampling from the prior
                    self.p0 = self.prior.sample_from_prior(self.n_hypers)
                    # Run MCMC sampling
                    self.p0, _, _ = self.sampler.run_mcmc(self.p0,
                                                          self.burnin_steps,
                                                          rstate0=self.rng)

                    self.burned = True

                # Start sampling
                pos, _, _ = self.sampler.run_mcmc(self.p0,
                                                  self.chain_length,
                                                  rstate0=self.rng)

                # Save the current position, it will be the startpoint in
                # the next iteration
                self.p0 = pos

                # Take the last samples from each walker set them back on a linear scale
                linear_theta = np.exp(self.sampler.chain[:, -1])
                self.hypers = linear_theta
                self.hypers[:, 1] = 1 / self.hypers[:, 1]
            else:
                # Optimize hyperparameters of the Bayesian linear regression
                p0 = self.prior.sample_from_prior(n_samples=1)
                res = optimize.fmin(self.negative_mll, p0)
                self.hypers = [[np.exp(res[0]), 1 / np.exp(res[1])]]
        else:

            self.hypers = [[self.alpha, self.beta]]

        logging.info("Hypers: %s" % self.hypers)
        self.models = []
        for sample in self.hypers:
            # Instantiate a model for each hyperparameter configuration
            model = BayesianLinearRegression(alpha=sample[0],
                                             beta=sample[1],
                                             basis_func=None)
            model.train(self.Theta, self.y[:, 0], do_optimize=False)

            self.models.append(model)

    def marginal_log_likelihood(self, theta):
        """
        Log likelihood of the data marginalised over the weights w. See chapter 3.5 of
        the book by Bishop of an derivation.

        Parameters
        ----------
        theta: np.array(2,)
            The hyperparameter alpha and beta on a log scale

        Returns
        -------
        float
            lnlikelihood + prior
        """
        if np.any(theta == np.inf):
            return -np.inf

        if np.any((-10 > theta) + (theta > 10)):
            return -np.inf

        alpha = np.exp(theta[0])
        beta = 1 / np.exp(theta[1])

        D = self.Theta.shape[1]
        N = self.Theta.shape[0]

        K = beta * np.dot(self.Theta.T, self.Theta)
        K += np.eye(self.Theta.shape[1]) * alpha
        K_inv = np.linalg.inv(K)
        m = beta * np.dot(K_inv, self.Theta.T)
        m = np.dot(m, self.y)

        mll = D / 2 * np.log(alpha)
        mll += N / 2 * np.log(beta)
        mll -= N / 2 * np.log(2 * np.pi)
        mll -= beta / 2. * np.linalg.norm(self.y - np.dot(self.Theta, m), 2)
        mll -= alpha / 2. * np.dot(m.T, m)
        mll -= 0.5 * np.log(np.linalg.det(K) + 1e-10)

        if np.any(np.isnan(mll)):
            return -1e25
        return mll

    def negative_mll(self, theta):
        """
        Returns the negative marginal log likelihood (for optimizing it with scipy).

        Parameters
        ----------
        theta: np.array(2,)
            The hyperparameter alpha and beta on a log scale

        Returns
        -------
        float
            negative lnlikelihood + prior
        """
        nll = -self.marginal_log_likelihood(theta)
        return nll

    def iterate_minibatches(self, inputs, targets, batchsize, shuffle=False):
        assert inputs.shape[0] == targets.shape[0], \
            "The number of training points is not the same"
        if shuffle:
            indices = np.arange(inputs.shape[0])
            self.rng.shuffle(indices)
        for start_idx in range(0, inputs.shape[0] - batchsize + 1, batchsize):
            if shuffle:
                excerpt = indices[start_idx:start_idx + batchsize]
            else:
                excerpt = slice(start_idx, start_idx + batchsize)
            yield inputs[excerpt], targets[excerpt]

    @BaseModel._check_shapes_predict
    def predict(self, X_test):
        r"""
        Returns the predictive mean and variance of the objective function at
        the given test points.

        Parameters
        ----------
        X_test: np.ndarray (N, D)
            N input test points

        Returns
        ----------
        np.array(N,)
            predictive mean
        np.array(N,)
            predictive variance

        """
        # Normalize inputs
        if self.normalize_input:
            X_, _, _ = zero_mean_unit_var_normalization(X_test, self.X_mean, self.X_std)
        else:
            X_ = X_test

        # Get features from the net

        theta = self.network.basis_funcs(torch.Tensor(X_)).data.numpy()

        # Marginalise predictions over hyperparameters of the BLR
        mu = np.zeros([len(self.models), X_test.shape[0]])
        var = np.zeros([len(self.models), X_test.shape[0]])

        for i, m in enumerate(self.models):
            mu[i], var[i] = m.predict(theta)

        # See the algorithm runtime prediction paper by Hutter et al
        # for the derivation of the total variance
        m = np.mean(mu, axis=0)
        v = np.mean(mu ** 2 + var, axis=0) - m ** 2

        # Clip negative variances and set them to the smallest
        # positive float value
        if v.shape[0] == 1:
            v = np.clip(v, np.finfo(v.dtype).eps, np.inf)
        else:
            v = np.clip(v, np.finfo(v.dtype).eps, np.inf)
            v[np.where((v < np.finfo(v.dtype).eps) & (v > -np.finfo(v.dtype).eps))] = 0

        if self.normalize_output:
            m = zero_mean_unit_var_denormalization(m, self.y_mean, self.y_std)
            v *= self.y_std ** 2

        return m, v

    def get_incumbent(self):
        """
        Returns the best observed point and its function value

        Returns
        ----------
        incumbent: ndarray (D,)
            current incumbent
        incumbent_value: ndarray (N,)
            the observed value of the incumbent
        """

        inc, inc_value = super(DNGO, self).get_incumbent()
        if self.normalize_input:
            inc = zero_mean_unit_var_denormalization(inc, self.X_mean, self.X_std)

        if self.normalize_output:
            inc_value = zero_mean_unit_var_denormalization(inc_value, self.y_mean, self.y_std)

        return inc, inc_value



import numpy as np
import torch
from torch.optim import Optimizer


# def decay(t, a, b , gamma):
#     return a * (b + t) ** (-gamma)
#
# def poly(base_lr, t, max_iter, power):
#     return base_lr * (1 - t / max_iter) ** power
#
# def const(*args):
#     return 1


class SGLD(Optimizer):
    """ Stochastic Gradient Langevin Dynamics Sampler
    """

    def __init__(self,
                 params,
                 lr: np.float64 = 1e-2,
                 scale_grad: np.float64 = 1) -> None:

        """ Set up a SGLD Optimizer.

        Parameters
        ----------
        params : iterable
            Parameters serving as optimization variable.
        lr : float, optional
            Base learning rate for this optimizer.
            Must be tuned to the specific function being minimized.
            Default: `1e-2`.
        """
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))

        # if lr_decay is None:
        #     self.lr_decay = const
        #     pass
        # elif lr_decay == "inv":
        #     final_lr_fraction = 1e-2
        #     degree = 2
        #     gamma = (np.power(1 / final_lr_fraction, 1. / degree) - 1) / (T - 1)
        #     self.lr_decay = lambda t: lr * np.power((1 + gamma * t), -degree)
        # else:
        #     self.lr_decay = lr_decay
        defaults = dict(
            lr=lr,
            scale_grad=scale_grad
        )
        super().__init__(params, defaults)

    def step(self, closure=None):
        loss = None

        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for parameter in group["params"]:

                if parameter.grad is None:
                    continue

                state = self.state[parameter]
                lr, scale_grad = group["lr"], group["scale_grad"]
                # the average gradient over the batch, i.e N/n sum_i g_theta_i + g_prior
                gradient = parameter.grad.data * scale_grad
                #  State initialization
                if len(state) == 0:
                    state["iteration"] = 0

                sigma = torch.sqrt(torch.from_numpy(np.array(lr, dtype=type(lr))))
                delta = (0.5 * lr * gradient +
                         sigma * torch.normal(mean=torch.zeros_like(gradient), std=torch.ones_like(gradient)))

                parameter.data.add_(-delta)
                state["iteration"] += 1
                state["sigma"] = sigma

        return loss



# vim: foldmethod=marker
from collections import OrderedDict
import typing
import numpy as np
import sympy
import torch
from torch.optim import Optimizer


def heaviside(x):
    return (x < 0).double() * torch.zeros_like(x) + (x >= 0).double() * torch.ones_like(x)

def maximum(a, b):
    return (a >= b).double() * a + (a < b).double() * b


def sympy_derivative(with_respect_to: str, tensor_names: typing.List[str], delta=1e-7):
    assert with_respect_to in tensor_names
    symbols = OrderedDict((
        (name, sympy.symbols(name))
        for name in tensor_names
    ))

    g2, p = symbols["g2"], symbols["p"]
    epsilon, mdecay, noise = symbols["epsilon"], symbols["mdecay"], symbols["noise"]
    grad, scale_grad = symbols["grad"], symbols["scale_grad"]
    random_sample = symbols["random_sample"]

    Minv = 1. / (sympy.sqrt(g2 + delta) + delta)

    epsilon_scaled = epsilon / sympy.sqrt(scale_grad)
    noise_scale = 2. * (epsilon_scaled ** 2) * mdecay * Minv - 2. * epsilon_scaled ** 3 * (Minv ** 2) * noise
    sigma = sympy.sqrt(sympy.Max(noise_scale, delta))

    sample_t = random_sample * sigma
    p_t = p - (epsilon ** 2) * Minv * grad - mdecay * p + sample_t

    derivative = sympy.diff(
        p_t, symbols[with_respect_to]
    )

    # Callable derivative function that computes d_theta d_`with respect to`
    return sympy.lambdify(
        args=tuple(symbols.values()),
        expr=derivative,
        modules={"sqrt": torch.sqrt, "Heaviside": heaviside, "Max": maximum}
    )


class SGHMCHD(Optimizer):
    name = "SGHMCHD"

    def __init__(self,
                 params,
                 hyper_lr: np.float64=1e-3,
                 lr: np.float64=1e-2,
                 num_burn_in_steps: int=3000,
                 noise: np.float64=0.,
                 mdecay: np.float64=0.05,
                 scale_grad: np.float64=1.) -> None:
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if num_burn_in_steps < 0:
            raise ValueError("Invalid num_burn_in_steps: {}".format(num_burn_in_steps))

        self.hyper_lr = hyper_lr
        self.num_burn_in_steps = num_burn_in_steps

        defaults = dict(
            lr=lr, scale_grad=float(scale_grad),
            mdecay=mdecay,
            noise=noise
        )
        super().__init__(params, defaults)

    def step(self, closure=None):
        loss = None

        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for parameter in group["params"]:

                if parameter.grad is None:
                    continue

                state = self.state[parameter]

                # initialize state

                if len(state) == 0:
                    state["iteration"] = 0
                    state["tau"] = torch.ones_like(parameter)
                    state["g"] = torch.ones_like(parameter)
                    state["v_hat"] = torch.ones_like(parameter)
                    state["momentum"] = torch.zeros_like(parameter)
                    # state["lr"] = torch.tensor(
                    #     torch.ones_like(parameter) * group["lr"], requires_grad=True
                    # )

                    state["lr"] = torch.ones_like(parameter) * group["lr"]
                    # state["lr"] = torch.tensor(
                    state["dxdh"] = torch.zeros_like(parameter)

                    state["dxdh_function"] = sympy_derivative(
                        with_respect_to="epsilon",
                        tensor_names=(
                            "g2", "p", "epsilon", "mdecay", "noise",
                            "grad", "scale_grad", "random_sample"
                        )
                    )

                    # import numpy as np
                    # from torch.optim import Adamax
                    # state["hyperoptimizer"] = Adamax(
                    #     params=(state["lr"],),
                    #     lr=self.hyper_lr)

                state["iteration"] += 1

                mdecay, noise, lr = group["mdecay"], group["noise"], state["lr"]
                scale_grad = torch.tensor(group["scale_grad"]).double()
                tau, g, v_hat = state["tau"], state["g"], state["v_hat"]
                momentum = state["momentum"]

                gradient = parameter.grad.data

                random_sample = torch.normal(mean=0., std=torch.ones_like(parameter)).double()

                # Update derivative of parameters with respect to hyperparameter
                torch_tensors = (
                    v_hat, momentum, lr, mdecay, noise,
                    gradient, scale_grad, random_sample

                )
                dxdh_t = state["dxdh_function"](*torch_tensors)  # hyper-gradient

                lr = lr - self.hyper_lr * gradient * dxdh_t
                state["lr"] = lr

                # state["hyperoptimizer"].zero_grad()
                # dxdlr = gradient * dxdh_t
                # state["lr"].grad = dxdlr
                # state["lr"].grad.data = dxdlr
                # state["hyperoptimizer"].step()
                # Update gradient of parameters with respect to hyperparameter.
                state["dxdh"].copy_(dxdh_t)

                r_t = 1. / (tau + 1.)
                minv_t = 1. / torch.sqrt(v_hat)

                #  burn-in
                if state["iteration"] <= self.num_burn_in_steps:
                    tau.add_(1. - tau * (g * g / v_hat))
                    g.add_(-g * r_t + r_t * gradient)
                    v_hat.add_(-v_hat * r_t + r_t * (gradient ** 2))

                lr_scaled = lr #/ torch.sqrt(scale_grad)

                noise_scale = (
                    2. * (lr_scaled ** 2) * mdecay * minv_t -
                    2. * (lr_scaled ** 3) * (minv_t ** 2) * noise -
                    (lr_scaled ** 4)
                )

                sigma = torch.sqrt(torch.clamp(noise_scale, min=1e-16))

                sample_t = random_sample * sigma

                momentum_t = momentum.add_(
                    - (lr ** 2) * minv_t * gradient - mdecay * momentum + sample_t)

                parameter.data.add_(momentum_t)

        return loss



import unittest
import numpy as np

from pybnn.util import normalization


class TestNormalization(unittest.TestCase):

    def test_zero_one_normalization(self):

        X = np.random.randn(100, 3)
        X_norm, lo, up = normalization.zero_one_normalization(X)

        assert X_norm.shape == X.shape
        assert np.min(X_norm) >= 0
        assert np.max(X_norm) <= 1
        assert lo.shape[0] == X.shape[1]
        assert up.shape[0] == X.shape[1]

    def test_zero_one_unnormalization(self):
        X_norm = np.random.rand(100, 3)
        lo = np.ones([3]) * -1
        up = np.ones([3])
        X = normalization.zero_one_denormalization(X_norm, lo, up)

        assert X_norm.shape == X.shape
        assert np.all(np.min(X, axis=0) >= lo)
        assert np.all(np.max(X, axis=0) <= up)

    def test_zero_mean_unit_var_normalization(self):
        X = np.random.rand(100, 3)
        X_norm, m, s = normalization.zero_mean_unit_var_normalization(X)

        np.testing.assert_almost_equal(np.mean(X_norm, axis=0), np.zeros(X_norm.shape[1]), decimal=1)
        np.testing.assert_almost_equal(np.var(X_norm, axis=0), np.ones(X_norm.shape[1]), decimal=1)

        assert X_norm.shape == X.shape
        assert m.shape[0] == X.shape[1]
        assert s.shape[0] == X.shape[1]

    def test_zero_one_unit_var_unnormalization(self):
        X_norm = np.random.randn(100, 3)
        m = np.ones(X_norm.shape[1]) * 3
        s = np.ones(X_norm.shape[1]) * 0.1
        X = normalization.zero_mean_unit_var_denormalization(X_norm, m, s)

        np.testing.assert_almost_equal(np.mean(X, axis=0), m, decimal=1)
        np.testing.assert_almost_equal(np.var(X, axis=0), s**2, decimal=1)

        assert X_norm.shape == X.shape

if __name__ == "__main__":
    unittest.main()



import torch
import torch.nn as nn
import numpy as np


class AppendLayer(nn.Module):
    def __init__(self, noise=1e-3, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log_var = nn.Parameter(torch.DoubleTensor(1, 1))

        nn.init.constant_(self.log_var, val=np.log(noise))

    def forward(self, x):
        return torch.cat((x, self.log_var * torch.ones_like(x)), dim=1)



import unittest

import numpy as np
from scipy.optimize import check_grad

from pybnn.bohamiann import Bohamiann


class TestBohamiann(unittest.TestCase):

    def setUp(self):
        self.X = np.random.rand(10, 3)
        self.y = np.sinc(self.X * 10 - 5).sum(axis=1)
        self.model = Bohamiann(normalize_input=True, normalize_output=True, use_double_precision=True)
        self.model.train(self.X, self.y, num_burn_in_steps=20, num_steps=100, keep_every=10)

    def test_predict(self):
        X_test = np.random.rand(10, self.X.shape[1])

        m, v = self.model.predict(X_test)

        assert len(m.shape) == 1
        assert m.shape[0] == X_test.shape[0]
        assert len(v.shape) == 1
        assert v.shape[0] == X_test.shape[0]

    def test_gradient_mean(self):
        X_test = np.random.rand(10, self.X.shape[1])

        def wrapper(x):
            return self.model.predict([x])[0]

        def wrapper_grad(x):
            return self.model.predictive_mean_gradient(x)

        grad = self.model.predictive_mean_gradient(X_test[0])
        assert grad.shape[0] == X_test.shape[1]

        for xi in X_test:
            err = check_grad(wrapper, wrapper_grad, xi, epsilon=1e-6)
            assert err < 1e-5

    def test_gradient_variance(self):
        X_test = np.random.rand(10, self.X.shape[1])

        def wrapper(x):
            v = self.model.predict([x])[1]
            return v

        def wrapper_grad(x):
            return self.model.predictive_variance_gradient(x)

        grad = self.model.predictive_variance_gradient(X_test[0])
        assert grad.shape[0] == X_test.shape[1]

        for xi in X_test:
            err = check_grad(wrapper, wrapper_grad, xi, epsilon=1e-6)
            assert err < 1e-5


class TestBohamiannSampler(unittest.TestCase):

    def test_sgld(self):
        self.X = np.random.rand(10, 3)
        self.y = np.sinc(self.X * 10 - 5).sum(axis=1)
        self.model = Bohamiann(normalize_input=True, normalize_output=True,
                               use_double_precision=True, sampling_method="sgld")
        self.model.train(self.X, self.y, num_burn_in_steps=20, num_steps=100, keep_every=10)

    def test_preconditioned_sgld(self):
        self.X = np.random.rand(10, 3)
        self.y = np.sinc(self.X * 10 - 5).sum(axis=1)
        self.model = Bohamiann(normalize_input=True, normalize_output=True,
                               use_double_precision=True, sampling_method="preconditioned_sgld")
        self.model.train(self.X, self.y, num_burn_in_steps=20, num_steps=100, keep_every=10)

    def test_sghmc(self):
        self.X = np.random.rand(10, 3)
        self.y = np.sinc(self.X * 10 - 5).sum(axis=1)
        self.model = Bohamiann(normalize_input=True, normalize_output=True,
                               use_double_precision=True, sampling_method="sghmc")
        self.model.train(self.X, self.y, num_burn_in_steps=20, num_steps=100, keep_every=10)

    def test_adaptive_sghmc(self):
        self.X = np.random.rand(10, 3)
        self.y = np.sinc(self.X * 10 - 5).sum(axis=1)
        self.model = Bohamiann(normalize_input=True, normalize_output=True,
                               use_double_precision=True, sampling_method="adaptive_sghmc")
        self.model.train(self.X, self.y, num_burn_in_steps=20, num_steps=100, keep_every=10)


if __name__ == "__main__":
    unittest.main()



from typing import Iterable
import numpy as np
import torch


def log_variance_prior(log_variance: torch.Tensor, mean: float = 1e-6, variance: float = 0.01) -> torch.Tensor:
    return torch.mean(
        torch.sum(
            ((-((log_variance - torch.log(torch.tensor(mean, dtype=log_variance.dtype))) ** 2)) /
             (2. * variance)) - 0.5 * torch.log(torch.tensor(variance, dtype=log_variance.dtype)),
            dim=1
        )
    )


def weight_prior(parameters: Iterable[torch.Tensor], dtype=np.float64, wdecay: float = 1.) -> torch.Tensor:

    num_parameters = 0
    log_likelihood = torch.from_numpy(np.array(0, dtype=dtype))
    for parameter in parameters:
        num_parameters += parameter.numel()
        log_likelihood += torch.sum(-wdecay * 0.5 * (parameter ** 2))

    return log_likelihood / num_parameters



import torch
import numpy as np
from torch.optim import Optimizer


class SGHMC(Optimizer):
    """ Stochastic Gradient Hamiltonian Monte-Carlo Sampler that uses a burn-in
        procedure to adapt its own hyperparameters during the initial stages
        of sampling.

        See [1] for more details on Stochastic Gradient Hamiltonian Monte-Carlo.

        [1] T. Chen, E. B. Fox, C. Guestrin
            In Proceedings of Machine Learning Research 32 (2014).\n
            `Stochastic Gradient Hamiltonian Monte Carlo <https://arxiv.org/pdf/1402.4102.pdf>`_
    """
    name = "AdaptiveSGHMC"

    def __init__(self,
                 params,
                 lr: float=1e-2,
                 mdecay: float=0.01,
                 wd: float=0.00002,
                 scale_grad: float=1.) -> None:
        """ Set up a SGHMC Optimizer.

        Parameters
        ----------
        params : iterable
            Parameters serving as optimization variable.
        lr: float, optional
            Base learning rate for this optimizer.
            Must be tuned to the specific function being minimized.
            Default: `1e-2`.
        mdecay:float, optional
            (Constant) momentum decay per time-step.
            Default: `0.05`.
        scale_grad: float, optional
            Value that is used to scale the magnitude of the noise used
            during sampling. In a typical batches-of-data setting this usually
            corresponds to the number of examples in the entire dataset.
            Default: `1.0`.

        """
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))

        defaults = dict(
            lr=lr, scale_grad=scale_grad,
            mdecay=mdecay,
            wd=wd
        )
        super().__init__(params, defaults)

    def step(self, closure=None):
        loss = None

        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for parameter in group["params"]:

                if parameter.grad is None:
                    continue

                state = self.state[parameter]

                if len(state) == 0:
                    state["iteration"] = 0
                    # state["momentum"] = torch.zeros_like(parameter)
                    state["momentum"] = torch.randn(parameter.size(), dtype=parameter.dtype)

                state["iteration"] += 1

                mdecay, lr, wd = group["mdecay"], group["lr"], group["wd"]
                scale_grad = group["scale_grad"]

                momentum = state["momentum"]
                gradient = parameter.grad.data * scale_grad

                # sigma = torch.sqrt(2 * lr * mdecay / scale_grad)
                sigma = torch.sqrt(torch.from_numpy(np.array(2 * lr * mdecay, dtype=type(lr))))
                sample_t = torch.normal(mean=torch.zeros_like(gradient), std=torch.ones_like(gradient) * sigma)
                # momentum_update = (1 - mdecay) * momentum - lr * gradient + sample_t #- lr * wd * parameter.data

                # momentum_t = momentum.add_(
                #     - lr * gradient - mdecay * momentum + sample_t
                # )

                # parameter.data.add_(momentum_t)
                parameter.data.add_(lr * momentum)
                momentum.add_(-lr * gradient - mdecay * momentum + sample_t)
        return loss



import unittest

import numpy as np

from pybnn.multi_task_bohamiann import MultiTaskBohamiann


class TestMTBohamiann(unittest.TestCase):

    def test_train_predict(self):

        def objective(x, task):
            if task == 0:
                y = 0.5 * np.sin(3 * x[0]) * 4 * (x[0] - 1) * (x[0])
            elif task == 1:
                y = 0.5 * np.sin(3 * x[0] + 1) * 4 * (x[0] - 1) * (x[0])
            elif task == 2:
                y = 0.5 * np.sin(3 * x[0] + 2) * 4 * (x[0] - 1) * (x[0])
            return y

        upper = np.ones(1) * 6

        X = np.random.rand(30, 1) * upper
        y_t0 = np.array([objective(xi, 0) for xi in X[:10]])
        y_t1 = np.array([objective(xi, 1) for xi in X[10:20]])
        y_t2 = np.array([objective(xi, 2) for xi in X[20:]])
        y = np.hstack((y_t0, y_t1, y_t2))

        t_idx = np.zeros([30])
        t_idx[10:20] = 1
        t_idx[20:] = 2

        X = np.append(X, t_idx[:, None], axis=1)

        model = MultiTaskBohamiann(n_tasks=3)

        model.train(X, y, num_steps=500, num_burn_in_steps=40, lr=1e-2)

        X_test = np.random.rand(5, 1) * upper
        X_test = np.append(X_test, np.ones([X_test.shape[0], 1]), axis=1)
        m, v = model.predict(X_test)

        assert len(m.shape) == 1
        assert m.shape[0] == X_test.shape[0]
        assert len(v.shape) == 1
        assert v.shape[0] == X_test.shape[0]


if __name__ == "__main__":
    unittest.main()



import emcee
import logging
import numpy as np

from scipy import optimize
from scipy import stats

from pybnn.base_model import BaseModel


def linear_basis_func(x):
    return np.append(x, np.ones([x.shape[0], 1]), axis=1)


def quadratic_basis_func(x):
    x = np.append(x ** 2, x, axis=1)
    return np.append(x, np.ones([x.shape[0], 1]), axis=1)


logger = logging.getLogger(__name__)


class Prior(object):

    def __init__(self, rng=None):
        if rng is None:
            self.rng = np.random.RandomState(np.random.randint(0, 10000))
        else:
            self.rng = rng

    def lnprob(self, theta):
        """
        Compute the log probability for theta = [log alpha, log beta]
        :param theta:
        :return: log p(theta)
        """
        lp = 0
        lp += stats.norm.pdf(theta[0], loc=0, scale=1)  # log alpha
        lp += stats.norm.pdf(theta[1], loc=0, scale=1)  # log sigma^2

        return lp

    def sample_from_prior(self, n_samples):
        p0 = np.zeros([n_samples, 2])

        # Log alpha
        p0[:, 0] = self.rng.normal(loc=0,
                                   scale=1,
                                   size=n_samples)

        # Log sigma^2
        p0[:, 1] = self.rng.normal(loc=-3,
                                   scale=1,
                                   size=n_samples)
        return p0


class BayesianLinearRegression(BaseModel):

    def __init__(self, alpha=1, beta=1000, basis_func=linear_basis_func,
                 prior=None, do_mcmc=True, n_hypers=20, chain_length=2000,
                 burnin_steps=2000, rng=None):
        """
        Implementation of Bayesian linear regression. See chapter 3.3 of the book
        "Pattern Recognition and Machine Learning" by Bishop for more details.

        Parameters
        ----------
        alpha: float
            Specifies the variance of the prior for the weights w
        beta : float
            Defines the inverse of the noise, i.e. beta = 1 / sigma^2
        basis_func : function
            Function handle to transfer the input with via basis functions
            (see the code above for an example)
        prior: Prior object
            Prior for alpha and beta. If set to None the default prior is used
        do_mcmc: bool
            If set to true different values for alpha and beta are sampled via MCMC from the marginal log likelihood
            Otherwise the marginal log likelihood is optimized with scipy fmin function
        n_hypers : int
            Number of samples for alpha and beta
        chain_length : int
            The chain length of the MCMC sampler
        burnin_steps: int
            The number of burnin steps before the sampling procedure starts
        rng: np.random.RandomState
            Random number generator
        """

        if rng is None:
            self.rng = np.random.RandomState(np.random.randint(0, 10000))
        else:
            self.rng = rng

        self.X = None
        self.y = None
        self.alpha = alpha
        self.beta = beta
        self.basis_func = basis_func
        if prior is None:
            self.prior = Prior(rng=self.rng)
        else:
            self.prior = prior
        self.do_mcmc = do_mcmc
        self.n_hypers = n_hypers
        self.chain_length = chain_length
        self.burned = False
        self.burnin_steps = burnin_steps
        self.models = None

    def marginal_log_likelihood(self, theta):
        """
        Log likelihood of the data marginalised over the weights w. See chapter 3.5 of
        the book by Bishop of an derivation.

        Parameters
        ----------
        theta: np.array(2,)
            The hyperparameter alpha and beta on a log scale

        Returns
        -------
        float
            lnlikelihood + prior
        """

        # Theta is on a log scale
        alpha = np.exp(theta[0])
        beta = 1 / np.exp(theta[1])

        D = self.X_transformed.shape[1]
        N = self.X_transformed.shape[0]

        A = beta * np.dot(self.X_transformed.T, self.X_transformed)
        A += np.eye(self.X_transformed.shape[1]) * alpha
        A_inv = np.linalg.inv(A)
        m = beta * np.dot(A_inv, self.X_transformed.T)
        m = np.dot(m, self.y)

        mll = D / 2 * np.log(alpha)
        mll += N / 2 * np.log(beta)
        mll -= N / 2 * np.log(2 * np.pi)
        mll -= beta / 2. * np.linalg.norm(self.y - np.dot(self.X_transformed, m), 2)
        mll -= alpha / 2. * np.dot(m.T, m)
        mll -= 0.5 * np.log(np.linalg.det(A))

        if self.prior is not None:
            mll += self.prior.lnprob(theta)

        return mll

    def negative_mll(self, theta):
        """
        Returns the negative marginal log likelihood (for optimizing it with scipy).

        Parameters
        ----------
        theta: np.array(2,)
            The hyperparameter alpha and beta on a log scale

        Returns
        -------
        float
            negative lnlikelihood + prior
        """
        return -self.marginal_log_likelihood(theta)

    @BaseModel._check_shapes_train
    def train(self, X, y, do_optimize=True):
        """
        First optimized the hyperparameters if do_optimize is True and then computes
        the posterior distribution of the weights. See chapter 3.3 of the book by Bishop
        for more details.

        Parameters
        ----------
        X: np.ndarray (N, D)
            Input data points. The dimensionality of X is (N, D),
            with N as the number of points and D is the number of features.
        y: np.ndarray (N,)
            The corresponding target values.
        do_optimize: boolean
            If set to true the hyperparameters are optimized otherwise
            the default hyperparameters are used.
        """

        self.X = X

        if self.basis_func is not None:
            self.X_transformed = self.basis_func(X)
        else:
            self.X_transformed = self.X

        self.y = y

        if do_optimize:
            if self.do_mcmc:
                sampler = emcee.EnsembleSampler(self.n_hypers, 2,
                                                self.marginal_log_likelihood)

                # Do a burn-in in the first iteration
                if not self.burned:
                    # Initialize the walkers by sampling from the prior
                    self.p0 = self.prior.sample_from_prior(self.n_hypers)

                    # Run MCMC sampling
                    self.p0, _, _ = sampler.run_mcmc(self.p0,
                                                     self.burnin_steps,
                                                     rstate0=self.rng)

                    self.burned = True

                # Start sampling
                pos, _, _ = sampler.run_mcmc(self.p0,
                                             self.chain_length,
                                             rstate0=self.rng)

                # Save the current position, it will be the start point in
                # the next iteration
                self.p0 = pos

                # Take the last samples from each walker
                self.hypers = np.exp(sampler.chain[:, -1])
            else:
                # Optimize hyperparameters of the Bayesian linear regression        
                res = optimize.fmin(self.negative_mll, self.rng.rand(2))
                self.hypers = [[np.exp(res[0]), np.exp(res[1])]]

        else:
            self.hypers = [[self.alpha, self.beta]]

        self.models = []
        for sample in self.hypers:
            alpha = sample[0]
            beta = sample[1]

            logger.debug("Alpha=%f ; Beta=%f" % (alpha, beta))

            S_inv = beta * np.dot(self.X_transformed.T, self.X_transformed)
            S_inv += np.eye(self.X_transformed.shape[1]) * alpha

            S = np.linalg.inv(S_inv)
            m = beta * np.dot(np.dot(S, self.X_transformed.T), self.y)

            self.models.append((m, S))

    @BaseModel._check_shapes_predict
    def predict(self, X_test):
        r"""
        Returns the predictive mean and variance of the objective function at
        the given test points.

        Parameters
        ----------
        X_test: np.ndarray (N, D)
            N input test points

        Returns
        ----------
        np.array(N,)
            predictive mean
        np.array(N,)
            predictive variance

        """
        if self.basis_func is not None:
            X_transformed = self.basis_func(X_test)
        else:
            X_transformed = X_test

        # Marginalise predictions over hyperparameters
        mu = np.zeros([len(self.hypers), X_transformed.shape[0]])
        var = np.zeros([len(self.hypers), X_transformed.shape[0]])

        for i, h in enumerate(self.hypers):
            mu[i] = np.dot(self.models[i][0].T, X_transformed.T)
            var[i] = 1. / h[1] + np.diag(np.dot(np.dot(X_transformed, self.models[i][1]), X_transformed.T))

        m = mu.mean(axis=0)
        v = var.mean(axis=0)
        # Clip negative variances and set them to the smallest
        # positive float value
        if v.shape[0] == 1:
            v = np.clip(v, np.finfo(v.dtype).eps, np.inf)
        else:
            v = np.clip(v, np.finfo(v.dtype).eps, np.inf)
            v[np.where((v < np.finfo(v.dtype).eps) & (v > -np.finfo(v.dtype).eps))] = 0

        return m, v



import numpy as np


def zero_one_normalization(X, lower=None, upper=None):

    if lower is None:
        lower = np.min(X, axis=0)
    if upper is None:
        upper = np.max(X, axis=0)

    X_normalized = np.true_divide((X - lower), (upper - lower))

    return X_normalized, lower, upper


def zero_one_denormalization(X_normalized, lower, upper):
    return lower + (upper - lower) * X_normalized


def zero_mean_unit_var_normalization(X, mean=None, std=None):
    if mean is None:
        mean = np.mean(X, axis=0)
    if std is None:
        std = np.std(X, axis=0)

    X_normalized = (X - mean) / std

    return X_normalized, mean, std


def zero_mean_unit_var_denormalization(X_normalized, mean, std):
    return X_normalized * std + mean



def infinite_dataloader(dataloader):
    """ Yield an unbounded amount of batches from a `torch.utils.data.DataLoader`.
    Parameters
    ----------
    dataloader : torch.utils.data.DataLoader
        Iterable yielding batches of data from a dataset of interest.
    """
    while True:
        for batch in dataloader:
            yield batch


from copy import deepcopy

import numpy as np
import torch
import torch.nn as nn
from pybnn.bohamiann import Bohamiann

"""
def vapor_pressure(t, a, b, c):
    a_ = a
    b_ = b / 10.
    c_ = c / 10.

    return torch.exp(-a_ - b_ / t - c_ * torch.log(t)) - torch.exp(-a_ - b_)


def pow_func(t, a, b):
    return a * (t ** b - 1)


def log_power(t, a, b, c):
    b_ = b / 10

    return 2 * a * (1 / (1 + torch.exp(-b_ * c)) - 1 / (1 + torch.exp(c * torch.log(t) - b_)))


def exponential(t, a, b):
    b_ = 10 * b
    return a * (torch.exp(-b_) - torch.exp(-b_ * t))


def hill_3(t, a, b, c):
    return a * (1 / (c ** b * t ** (-b) + 1) - a / (c ** b + 1))

def bf_layer(theta, t):
    a = theta[:, 0]
    b = theta[:, 1]
    c = theta[:, 2]
    y_a = vapor_pressure(t, a, b, c)

    a = theta[:, 3]
    b = theta[:, 4]
    y_b = pow_func(t, a, b)

    a = theta[:, 5]
    b = theta[:, 6]
    c = theta[:, 7]
    y_c = log_power(t, a, b, c)

    a = theta[:, 8]
    b = theta[:, 9]
    y_d = exponential(t, a, b)

    a = theta[:, 10]
    b = theta[:, 11]
    c = theta[:, 12]
    y_e = hill_3(t, a, b, c)

    return torch.stack([y_a, y_b, y_c, y_d, y_e], dim=1)

"""


def vapor_pressure(x, a, b, c, *args):
    a_ = -a
    b_ = -b / 10.
    c_ = -c / 10.

    return (torch.exp(a_ + b_ / (x + 1e-5) + c_ * torch.log(x + 1e-5)) - torch.exp(a_ + b_))


def log_log_linear(x, a, b, *args):
    return torch.log(1 + a * torch.log(1 + x) + b)


def hill_3(x, a, b, c, *args):
    return a * (1. / ((c / x) ** b + 1.) - 1. / (c ** b + 1.))


def pow_func(x, a, b, *args):
    return a * ((x) ** b - 1)


def log_power(x, a, b, c, *args):
    b_n = b * 0.1
    return 2 * a * (1 / (1. + torch.exp(-c * b_n)) - 1 / (1. + (x / torch.exp(b_n)) ** c))


def pow_4(x, a, b, c, *args):
    a_n = a
    b_n = b / 10.
    c_n = c / 10.
    d = 1 + (a_n + b_n) ** (-c_n)
    return d - (a_n * x + b_n) ** (-c_n)


def mmf(x, a, b, c, d, *args):
    return a - (a - b) / (1 + (c * x) ** d)


def exp_4(x, a, b, c, d, *args):
    f_max = 1 - np.exp(-1)
    f_min = 0 - np.exp(1)
    return ((d - torch.exp(-a * x ** c + b)) - f_min) / (f_max - f_min)


def janoschek(x, a, b, c, d, *args):
    # c_n = c ** 2
    c_n = c
    return a - (a - b) * torch.exp(-c_n * x ** d)


def weibull(x, a, b, c, d, *args):
    # c_n = c ** 2
    return a - (a - b) * torch.exp(-(c * x) ** d)


def ilog_2(x, a, *args):
    a_n = a * 0.1
    b = a_n / np.log(2.1) + 1
    return b - a_n / torch.log((x + 1.1))


def linear(x, bias=0, a=1, *args):
    return a * x + bias


def quadratic(x, bias=0, a=0.1, b=0.1, *args):
    return a * (x ** 2) + b * x + bias


def exponential(x, a, b, *args):
    return b * (- torch.exp(-10 * a * x) + torch.exp(-10 * a))


def bf_layer(theta, t):
    y_a = vapor_pressure(t, theta[:, 0], theta[:, 1], theta[:, 2])
    y_b = log_log_linear(t, theta[:, 3], theta[:, 4])

    y_c = hill_3(t, theta[:, 5], theta[:, 6], theta[:, 7])

    y_d = pow_func(t, theta[:, 8], theta[:, 9])

    y_e = log_power(t, theta[:, 10], theta[:, 11], theta[:, 12])

    y_f = pow_4(t, theta[:, 13], theta[:, 14], theta[:, 15])

    y_g = mmf(t, theta[:, 16], theta[:, 17], theta[:, 18], theta[:, 19])

    y_h = exp_4(t, theta[:, 20], theta[:, 21], theta[:, 22], theta[:, 23])

    y_i = janoschek(t, theta[:, 24], theta[:, 25], theta[:, 26], theta[:, 27])

    y_j = weibull(t, theta[:, 28], theta[:, 29], theta[:, 30], theta[:, 31])

    y_k = ilog_2(t, theta[:, 32])

    y_l = linear(t, theta[:, 33], theta[:, 34])

    y_m = quadratic(t, theta[:, 35], theta[:, 36], theta[:, 37])

    y_n = exponential(t, theta[:, 38], theta[:, 39])

    return torch.stack([y_a, y_b, y_c, y_d, y_e, y_f, y_g, y_h, y_i, y_j, y_k, y_l, y_m, y_n], dim=1)


class AppendLayer(nn.Module):
    def __init__(self, bias=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if bias:
            self.bias = nn.Parameter(torch.DoubleTensor(1, 1))
        else:
            self.register_parameter('bias', None)

    def forward(self, x):
        return torch.cat((x, self.bias * torch.ones_like(x)), dim=1)


def get_lc_net_architecture(input_dimensionality: int) -> torch.nn.Module:
    class Architecture(nn.Module):
        def __init__(self, n_inputs, n_hidden=50):
            super(Architecture, self).__init__()
            self.fc1 = nn.Linear(n_inputs - 1, n_hidden)
            self.fc2 = nn.Linear(n_hidden, n_hidden)
            self.fc3 = nn.Linear(n_hidden, n_hidden)
            self.theta_layer = nn.Linear(n_hidden, 40)
            self.weight_layer = nn.Linear(n_hidden, 14)
            self.asymptotic_layer = nn.Linear(n_hidden, 1)
            # self.sigma_layer = nn.Linear(n_hidden, 1)
            self.sigma_layer = AppendLayer()

        def forward(self, input):
            x = input[:, :-1]
            t = input[:, -1]
            x = torch.tanh(self.fc1(x))
            x = torch.tanh(self.fc2(x))
            x = torch.tanh(self.fc3(x))
            theta = torch.sigmoid(self.theta_layer(x))

            bf = bf_layer(theta, t)
            weights = torch.softmax(self.weight_layer(x), -1)
            residual = torch.sum(bf * weights, dim=(1,), keepdim=True)

            asymptotic = torch.sigmoid(self.asymptotic_layer(x))

            mean = residual + asymptotic

            return self.sigma_layer(mean)
            # std = torch.sigmoid(self.sigma_layer(x))

            # return torch.cat((mean, std), dim=1)

    return Architecture(n_inputs=input_dimensionality)


class LCNet(Bohamiann):
    def __init__(self,
                 metrics=(nn.MSELoss,)
                 ) -> None:
        super(LCNet, self).__init__(get_network=get_lc_net_architecture,
                                    normalize_input=True,
                                    normalize_output=False,
                                    metrics=metrics)

    def normalize_input(self, x, m=None, s=None):
        if m is None:
            m = np.mean(x, axis=0)
        if s is None:
            s = np.std(x, axis=0)

        x_norm = deepcopy(x)
        x_norm[:, :-1] = (x[:, :-1] - m[:-1]) / s[:-1]

        return x_norm, m, s



import torch

from torch.optim import Optimizer


class AdaptiveSGHMC(Optimizer):
    """ Stochastic Gradient Hamiltonian Monte-Carlo Sampler that uses a burn-in
        procedure to adapt its own hyperparameters during the initial stages
        of sampling.

        See [1] for more details on this burn-in procedure.\n
        See [2] for more details on Stochastic Gradient Hamiltonian Monte-Carlo.

        [1] J. T. Springenberg, A. Klein, S. Falkner, F. Hutter
            In Advances in Neural Information Processing Systems 29 (2016).\n
            `Bayesian Optimization with Robust Bayesian Neural Networks. <http://aad.informatik.uni-freiburg.de/papers/16-NIPS-BOHamiANN.pdf>`_
        [2] T. Chen, E. B. Fox, C. Guestrin
            In Proceedings of Machine Learning Research 32 (2014).\n
            `Stochastic Gradient Hamiltonian Monte Carlo <https://arxiv.org/pdf/1402.4102.pdf>`_
    """

    def __init__(self,
                 params,
                 lr: float = 1e-2,
                 num_burn_in_steps: int = 3000,
                 epsilon: float = 1e-16,
                 mdecay: float = 0.05,
                 scale_grad: float = 1.) -> None:
        """ Set up a SGHMC Optimizer.

        Parameters
        ----------
        params : iterable
            Parameters serving as optimization variable.
        lr: float, optional
            Base learning rate for this optimizer.
            Must be tuned to the specific function being minimized.
            Default: `1e-2`.
        num_burn_in_steps: int, optional
            Number of burn-in steps to perform. In each burn-in step, this
            sampler will adapt its own internal parameters to decrease its error.
            Set to `0` to turn scale adaption off.
            Default: `3000`.
        epsilon: float, optional
            (Constant) per-parameter epsilon level.
            Default: `0.`.
        mdecay:float, optional
            (Constant) momentum decay per time-step.
            Default: `0.05`.
        scale_grad: float, optional
            Value that is used to scale the magnitude of the epsilon used
            during sampling. In a typical batches-of-data setting this usually
            corresponds to the number of examples in the entire dataset.
            Default: `1.0`.

        """
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if num_burn_in_steps < 0:
            raise ValueError("Invalid num_burn_in_steps: {}".format(num_burn_in_steps))

        defaults = dict(
            lr=lr, scale_grad=float(scale_grad),
            num_burn_in_steps=num_burn_in_steps,
            mdecay=mdecay,
            epsilon=epsilon
        )
        super().__init__(params, defaults)

    def step(self, closure=None):
        loss = None

        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for parameter in group["params"]:

                if parameter.grad is None:
                    continue

                state = self.state[parameter]

                if len(state) == 0:
                    state["iteration"] = 0
                    state["tau"] = torch.ones_like(parameter)
                    state["g"] = torch.ones_like(parameter)
                    state["v_hat"] = torch.ones_like(parameter)
                    state["momentum"] = torch.zeros_like(parameter)
                state["iteration"] += 1

                mdecay, epsilon, lr = group["mdecay"], group["epsilon"], group["lr"]
                scale_grad = torch.tensor(group["scale_grad"], dtype=parameter.dtype)
                tau, g, v_hat = state["tau"], state["g"], state["v_hat"]

                momentum = state["momentum"]
                gradient = parameter.grad.data * scale_grad

                tau_inv = 1. / (tau + 1.)

                # update parameters during burn-in
                if state["iteration"] <= group["num_burn_in_steps"]:
                    tau.add_(- tau * (
                            g * g / (v_hat + epsilon)) + 1)  # specifies the moving average window, see Eq 9 in [1] left
                    g.add_(-g * tau_inv + tau_inv * gradient)  # average gradient see Eq 9 in [1] right
                    v_hat.add_(-v_hat * tau_inv + tau_inv * (gradient ** 2))  # gradient variance see Eq 8 in [1]

                minv_t = 1. / (torch.sqrt(v_hat) + epsilon)  # preconditioner

                epsilon_var = (2. * (lr ** 2) * mdecay * minv_t - (lr ** 4))

                # sample random epsilon
                sigma = torch.sqrt(torch.clamp(epsilon_var, min=1e-16))
                sample_t = torch.normal(mean=torch.zeros_like(gradient), std=torch.ones_like(gradient) * sigma)

                # update momentum (Eq 10 right in [1])
                momentum.add_(
                    - (lr ** 2) * minv_t * gradient - mdecay * momentum + sample_t
                )

                # update theta (Eq 10 left in [1])
                parameter.data.add_(momentum)

        return loss



import unittest

import numpy as np

from pybnn.lcnet import LCNet


class TestLCNet(unittest.TestCase):

    def test_train_predict(self):

        def toy_example(t, a, b):
            return (10 + a * np.log(b * t)) / 10. + 10e-3 * np.random.rand()

        observed = 20
        N = 5
        n_epochs = 10
        observed_t = int(n_epochs * (observed / 100.))

        t_idx = np.arange(1, observed_t + 1) / n_epochs
        t_grid = np.arange(1, n_epochs + 1) / n_epochs

        configs = np.random.rand(N, 2)
        learning_curves = [toy_example(t_grid, configs[i, 0], configs[i, 1]) for i in range(N)]

        X_train = None
        y_train = None
        X_test = None
        y_test = None

        for i in range(N):

            x = np.repeat(configs[i, None, :], t_idx.shape[0], axis=0)
            x = np.concatenate((x, t_idx[:, None]), axis=1)

            x_test = np.concatenate((configs[i, None, :], np.array([[1]])), axis=1)

            lc = learning_curves[i][:observed_t]
            lc_test = np.array([learning_curves[i][-1]])

            if X_train is None:
                X_train = x
                y_train = lc
                X_test = x_test
                y_test = lc_test
            else:
                X_train = np.concatenate((X_train, x), 0)
                y_train = np.concatenate((y_train, lc), 0)
                X_test = np.concatenate((X_test, x_test), 0)
                y_test = np.concatenate((y_test, lc_test), 0)

        print(X_train.shape)
        model = LCNet()

        model.train(X_train, y_train, num_steps=500, num_burn_in_steps=40, lr=1e-2)

        m, v = model.predict(X_test)

        assert len(m.shape) == 1
        assert m.shape[0] == X_test.shape[0]
        assert len(v.shape) == 1
        assert v.shape[0] == X_test.shape[0]


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from pybnn.dngo import DNGO


class TestDNGO(unittest.TestCase):

    def setUp(self):
        self.X = np.random.rand(10, 3)
        self.y = np.sinc(self.X * 10 - 5).sum(axis=1)

    def test_mcmc(self):
        model = DNGO(num_epochs=10, burnin_steps=10, chain_length=20, do_mcmc=True)
        model.train(self.X, self.y)

        X_test = np.random.rand(10, self.X.shape[1])

        m, v = model.predict(X_test)

        assert len(m.shape) == 1
        assert m.shape[0] == X_test.shape[0]
        assert len(v.shape) == 1
        assert v.shape[0] == X_test.shape[0]

    def test_ml(self):
        model = DNGO(num_epochs=10, do_mcmc=False)
        model.train(self.X, self.y)

        X_test = np.random.rand(10, self.X.shape[1])

        m, v = model.predict(X_test)

        assert len(m.shape) == 1
        assert m.shape[0] == X_test.shape[0]
        assert len(v.shape) == 1
        assert v.shape[0] == X_test.shape[0]

    def test_without_normalization(self):
        model = DNGO(num_epochs=10, do_mcmc=False, normalize_output=False, normalize_input=False)
        model.train(self.X, self.y)

        X_test = np.random.rand(10, self.X.shape[1])

        m, v = model.predict(X_test)

        assert len(m.shape) == 1
        assert m.shape[0] == X_test.shape[0]
        assert len(v.shape) == 1
        assert v.shape[0] == X_test.shape[0]

    def test_incumbent(self):
        model = DNGO(num_epochs=10, do_mcmc=False)
        model.train(self.X, self.y)

        x_star, y_star = model.get_incumbent()

        b = np.argmin(self.y)

        assert np.all(np.isclose(x_star, self.X[b]))
        assert np.all(np.isclose(y_star, self.y[b]))


if __name__ == "__main__":
    unittest.main()



import logging
import time
import typing
from itertools import islice

import numpy as np
import torch
import torch.nn as nn
import torch.utils.data as data_utils
from scipy.stats import norm

from pybnn.base_model import BaseModel
from pybnn.priors import weight_prior, log_variance_prior
from pybnn.sampler import AdaptiveSGHMC, SGLD, SGHMC, PreconditionedSGLD
from pybnn.util.infinite_dataloader import infinite_dataloader
from pybnn.util.layers import AppendLayer
from pybnn.util.normalization import zero_mean_unit_var_denormalization, zero_mean_unit_var_normalization


def get_default_network(input_dimensionality: int) -> torch.nn.Module:
    class Architecture(torch.nn.Module):
        def __init__(self, n_inputs, n_hidden=50):
            super(Architecture, self).__init__()
            self.fc1 = torch.nn.Linear(n_inputs, n_hidden)
            self.fc2 = torch.nn.Linear(n_hidden, n_hidden)
            self.fc3 = torch.nn.Linear(n_hidden, 1)
            self.log_std = AppendLayer(noise=1e-3)

        def forward(self, input):
            x = torch.tanh(self.fc1(input))
            x = torch.tanh(self.fc2(x))
            x = self.fc3(x)
            return self.log_std(x)

    return Architecture(n_inputs=input_dimensionality)


def nll(input: torch.Tensor, target: torch.Tensor):
    """
    computes the average negative log-likelihood (Gaussian)

    :param input: mean and variance predictions of the networks
    :param target: target values
    :return: negative log-likelihood
    """
    batch_size = input.size(0)

    prediction_mean = input[:, 0].view((-1, 1))
    log_prediction_variance = input[:, 1].view((-1, 1))
    prediction_variance_inverse = 1. / (torch.exp(log_prediction_variance) + 1e-16)

    mean_squared_error = (target.view(-1, 1) - prediction_mean) ** 2

    log_likelihood = torch.sum(
        torch.sum(-mean_squared_error * (0.5 * prediction_variance_inverse) - 0.5 * log_prediction_variance, dim=1))

    log_likelihood = log_likelihood / batch_size

    return -log_likelihood


class Bohamiann(BaseModel):
    def __init__(self,
                 get_network=get_default_network,
                 normalize_input: bool = True,
                 normalize_output: bool = True,
                 sampling_method: str = "adaptive_sghmc",
                 use_double_precision: bool = True,
                 metrics=(nn.MSELoss,),
                 likelihood_function=nll,
                 print_every_n_steps=100,
                 ) -> None:
        """

        Bayesian Neural Networks use Bayesian methods to estimate the posterior
        distribution of a neural network's weights. This allows to also
        predict uncertainties for test points and thus makes Bayesian Neural
        Networks suitable for Bayesian optimization.
        This module uses stochastic gradient MCMC methods to sample
        from the posterior distribution.

        See [1] for more details.

        [1] J. T. Springenberg, A. Klein, S. Falkner, F. Hutter
            Bayesian Optimization with Robust Bayesian Neural Networks.
            In Advances in Neural Information Processing Systems 29 (2016).

        :param get_network: function handle that returns the archtiecture
        :param normalize_input: defines whether to normalize the inputs
        :param normalize_output: defines whether to normalize the outputs
        :param sampling_method: specifies the sampling strategy,
        options: {sgld, sghmc, adaptive_sghmc, preconditioned_sgld}
        :param use_double_precision: defines whether to use double or float precisions
        :param metrics: metrics to evaluate
        :param likelihood_function: function handle that computes the training loss
        :param print_every_n_steps: defines after how many the current loss is printed
        """
        self.print_every_n_steps = print_every_n_steps
        self.metrics = metrics
        self.do_normalize_input = normalize_input
        self.do_normalize_output = normalize_output
        self.get_network = get_network
        self.is_trained = False
        self.use_double_precision = use_double_precision
        self.sampling_method = sampling_method
        self.sampled_weights = []  # type: typing.List[typing.Tuple[np.ndarray]]
        self.likelihood_function = likelihood_function
        self.sampler = None

    @property
    def network_weights(self) -> tuple:
        """
        Extract current network weight values as `np.ndarray`.

        :return: Tuple containing current network weight values
        """
        return tuple(
            np.asarray(torch.tensor(parameter.data).numpy())
            for parameter in self.model.parameters()
        )

    @network_weights.setter
    def network_weights(self, weights: typing.List[np.ndarray]) -> None:
        """
        Assign new `weights` to our neural networks parameters.

        :param weights: List of weight values to assign.
            Individual list elements must have shapes that match
            the network parameters with the same index in `self.network_weights`.
        """
        logging.debug("Assigning new network weights")
        for parameter, sample in zip(self.model.parameters(), weights):
            parameter.copy_(torch.from_numpy(sample))

    def train(self, x_train: np.ndarray, y_train: np.ndarray,
              num_steps: int = 13000,
              keep_every: int = 100,
              num_burn_in_steps: int = 3000,
              lr: float = 1e-5,
              batch_size=20,
              epsilon: float = 1e-10,
              mdecay: float = 0.05,
              continue_training: bool = False,
              verbose: bool = False,
              **kwargs):

        """
        Train a BNN using input datapoints `x_train` with corresponding targets `y_train`.

        :param x_train: input training datapoints.
        :param y_train: input training targets.
        :param num_steps: Number of sampling steps to perform after burn-in is finished.
            In total, `num_steps // keep_every` network weights will be sampled.
        :param keep_every: Number of sampling steps (after burn-in) to perform before keeping a sample.
            In total, `num_steps // keep_every` network weights will be sampled.
        :param num_burn_in_steps: Number of burn-in steps to perform.
            This value is passed to the given `optimizer` if it supports special
            burn-in specific behavior.
            Networks sampled during burn-in are discarded.
        :param lr: learning rate
        :param batch_size: batch size
        :param epsilon: epsilon for numerical stability
        :param mdecay: momemtum decay
        :param continue_training: defines whether we want to continue from the last training run
        :param verbose: verbose output
        """
        logging.debug("Training started.")
        start_time = time.time()

        num_datapoints, input_dimensionality = x_train.shape
        logging.debug(
            "Processing %d training datapoints "
            " with % dimensions each." % (num_datapoints, input_dimensionality)
        )
        assert batch_size >= 1, "Invalid batch size. Batches must contain at least a single sample."
        assert len(y_train.shape) == 1 or (len(y_train.shape) == 2 and y_train.shape[
            1] == 1), "Targets need to be in vector format, i.e (N,) or (N,1)"

        if x_train.shape[0] < batch_size:
            logging.warning("Not enough datapoints to form a batch. Use all datapoints in each batch")
            batch_size = x_train.shape[0]

        self.X = x_train
        if len(y_train.shape) == 2:
            self.y = y_train[:, 0]
        else:
            self.y = y_train

        if self.do_normalize_input:
            logging.debug(
                "Normalizing training datapoints to "
                " zero mean and unit variance."
            )
            x_train_, self.x_mean, self.x_std = self.normalize_input(x_train)
            if self.use_double_precision:
                x_train_ = torch.from_numpy(x_train_).double()
            else:
                x_train_ = torch.from_numpy(x_train_).float()
        else:
            if self.use_double_precision:
                x_train_ = torch.from_numpy(x_train).double()
            else:
                x_train_ = torch.from_numpy(x_train).float()

        if self.do_normalize_output:
            logging.debug("Normalizing training labels to zero mean and unit variance.")
            y_train_, self.y_mean, self.y_std = self.normalize_output(y_train)

            if self.use_double_precision:
                y_train_ = torch.from_numpy(y_train_).double()
            else:
                y_train_ = torch.from_numpy(y_train_).float()
        else:
            if self.use_double_precision:
                y_train_ = torch.from_numpy(y_train).double()
            else:
                y_train_ = torch.from_numpy(y_train).float()

        train_loader = infinite_dataloader(
            data_utils.DataLoader(
                data_utils.TensorDataset(x_train_, y_train_),
                batch_size=batch_size,
                shuffle=True
            )
        )

        if self.use_double_precision:
            dtype = np.float64
        else:
            dtype = np.float32

        if not continue_training:
            logging.debug("Clearing list of sampled weights.")

            self.sampled_weights.clear()
            if self.use_double_precision:
                self.model = self.get_network(input_dimensionality=input_dimensionality).double()
            else:
                self.model = self.get_network(input_dimensionality=input_dimensionality).float()

            if self.sampling_method == "adaptive_sghmc":
                self.sampler = AdaptiveSGHMC(self.model.parameters(),
                                             scale_grad=dtype(num_datapoints),
                                             num_burn_in_steps=num_burn_in_steps,
                                             lr=dtype(lr),
                                             mdecay=dtype(mdecay),
                                             epsilon=dtype(epsilon))
            elif self.sampling_method == "sgld":
                self.sampler = SGLD(self.model.parameters(),
                                    lr=dtype(lr),
                                    scale_grad=num_datapoints)
            elif self.sampling_method == "preconditioned_sgld":
                self.sampler = PreconditionedSGLD(self.model.parameters(),
                                                  lr=dtype(lr),
                                                  num_train_points=num_datapoints)
            elif self.sampling_method == "sghmc":
                self.sampler = SGHMC(self.model.parameters(),
                                     scale_grad=dtype(num_datapoints),
                                     mdecay=dtype(mdecay),
                                     lr=dtype(lr))

        batch_generator = islice(enumerate(train_loader), num_steps)

        for step, (x_batch, y_batch) in batch_generator:
            self.sampler.zero_grad()
            loss = self.likelihood_function(input=self.model(x_batch), target=y_batch)
            # Add prior. Note the gradient is computed by: g_prior + N/n sum_i grad_theta_xi see Eq 4
            # in Welling and Whye The 2011. Because of that we divide here by N=num of datapoints since
            # in the sample we rescale the gradient by N again
            loss -= log_variance_prior(self.model(x_batch)[:, 1].view((-1, 1))) / num_datapoints
            loss -= weight_prior(self.model.parameters(), dtype=dtype) / num_datapoints
            loss.backward()
            self.sampler.step()

            if verbose and step > 0 and step % self.print_every_n_steps == 0:

                # compute the training performance of the ensemble
                if len(self.sampled_weights) > 1:
                    mu, var = self.predict(x_train)
                    total_nll = -np.mean(norm.logpdf(y_train, loc=mu, scale=np.sqrt(var)))
                    total_mse = np.mean((y_train - mu) ** 2)
                # in case we do not have an ensemble we compute the performance of the last weight sample
                else:
                    f = self.model(x_train_)

                    if self.do_normalize_output:
                        mu = zero_mean_unit_var_denormalization(f[:, 0], self.y_mean, self.y_std).data.numpy()
                        var = torch.exp(f[:, 1]) * self.y_std ** 2
                        var = var.data.numpy()
                    else:
                        mu = f[:, 0].data.numpy()
                        var = f[:, 1].data.numpy()
                    total_nll = -np.mean(norm.logpdf(y_train, loc=mu, scale=np.sqrt(var)))
                    total_mse = np.mean((y_train - mu) ** 2)

                t = time.time() - start_time

                if step < num_burn_in_steps:
                    print("Step {:8d} : NLL = {:11.4e} MSE = {:.4e} "
                          "Time = {:5.2f}".format(step, float(total_nll),
                                                  float(total_mse), t))

                if step > num_burn_in_steps:
                    print("Step {:8d} : NLL = {:11.4e} MSE = {:.4e} "
                          "Samples= {} Time = {:5.2f}".format(step,
                                                              float(total_nll),
                                                              float(total_mse),
                                                              len(self.sampled_weights), t))

            if step > num_burn_in_steps and (step - num_burn_in_steps) % keep_every == 0:
                weights = self.network_weights

                self.sampled_weights.append(weights)

        self.is_trained = True

    def train_and_evaluate(self, x_train: np.ndarray, y_train: np.ndarray,
                           x_valid: np.ndarray, y_valid: np.ndarray,
                           num_steps: int = 13000,
                           validate_every_n_steps=1000,
                           keep_every: int = 100,
                           num_burn_in_steps: int = 3000,
                           lr: float = 1e-2,
                           epsilon: float = 1e-10,
                           batch_size: int = 20,
                           mdecay: float = 0.05,
                           verbose=False):
        """
        Train and validates the neural network

        :param x_train: input training datapoints.
        :param y_train: input training targets.
        :param x_valid: validation data points
        :param y_valid: valdiation targets
        :param num_steps:  Number of sampling steps to perform after burn-in is finished.
            In total, `num_steps // keep_every` network weights will be sampled.
        :param validate_every_n_steps:
        :param keep_every: Number of sampling steps (after burn-in) to perform before keeping a sample.
            In total, `num_steps // keep_every` network weights will be sampled.
        :param num_burn_in_steps: Number of burn-in steps to perform.
            This value is passed to the given `optimizer` if it supports special
            burn-in specific behavior.
            Networks sampled during burn-in are discarded.
        :param lr: learning rate
        :param batch_size: batch size
        :param epsilon: epsilon for numerical stability
        :param mdecay: momemtum decay
        :param verbose: verbose output

        """
        assert batch_size >= 1, "Invalid batch size. Batches must contain at least a single sample."

        if x_train.shape[0] < batch_size:
            logging.warning("Not enough datapoints to form a batch. Use all datapoints in each batch")
            batch_size = x_train.shape[0]

        # burn-in
        self.train(x_train, y_train, num_burn_in_steps=num_burn_in_steps, num_steps=num_burn_in_steps,
                   lr=lr, epsilon=epsilon, mdecay=mdecay, verbose=verbose)

        learning_curve_mse = []
        learning_curve_ll = []
        n_steps = []
        for i in range(num_steps // validate_every_n_steps):
            self.train(x_train, y_train, num_burn_in_steps=0, num_steps=validate_every_n_steps,
                       lr=lr, epsilon=epsilon, mdecay=mdecay, verbose=verbose, keep_every=keep_every,
                       continue_training=True, batch_size=batch_size)

            mu, var = self.predict(x_valid)

            ll = np.mean(norm.logpdf(y_valid, loc=mu, scale=np.sqrt(var)))
            mse = np.mean((y_valid - mu) ** 2)
            step = num_burn_in_steps + (i + 1) * validate_every_n_steps

            learning_curve_ll.append(ll)
            learning_curve_mse.append(mse)
            n_steps.append(step)

            if verbose:
                print("Validate : NLL = {:11.4e} MSE = {:.4e}".format(-ll, mse))

        return n_steps, learning_curve_ll, learning_curve_mse

    def normalize_input(self, x, m=None, s=None):
        """
        Normalizes input

        :param x: data
        :param m: mean
        :param s: standard deviation
        :return: normalized input
        """

        return zero_mean_unit_var_normalization(x, m, s)

    def normalize_output(self, x, m=None, s=None):
        """
        Normalizes output

        :param x: targets
        :param m: mean
        :param s: standard deviation
        :return: normalized targets
        """
        return zero_mean_unit_var_normalization(x, m, s)

    def predict(self, x_test: np.ndarray, return_individual_predictions: bool = False):
        """
        Predicts mean and variance for the given test point

        :param x_test: test datapoint
        :param return_individual_predictions: if True also the predictions of the individual models are returned
        :return: mean and variance
        """
        x_test_ = np.asarray(x_test)

        if self.do_normalize_input:
            x_test_, *_ = self.normalize_input(x_test_, self.x_mean, self.x_std)

        def network_predict(x_test_, weights):
            with torch.no_grad():
                self.network_weights = weights
                if self.use_double_precision:
                    return self.model(torch.from_numpy(x_test_).double()).numpy()
                else:
                    return self.model(torch.from_numpy(x_test_).float()).numpy()

        logging.debug("Predicting with %d networks." % len(self.sampled_weights))
        network_outputs = np.array([
            network_predict(x_test_, weights=weights)
            for weights in self.sampled_weights
        ])

        mean_prediction = np.mean(network_outputs[:, :, 0], axis=0)
        # variance_prediction = np.mean((network_outputs[:, :, 0] - mean_prediction) ** 2, axis=0)
        # Total variance
        variance_prediction = np.mean((network_outputs[:, :, 0] - mean_prediction) ** 2
                                      + np.exp(network_outputs[:, :, 1]), axis=0)

        if self.do_normalize_output:

            mean_prediction = zero_mean_unit_var_denormalization(
                mean_prediction, self.y_mean, self.y_std
            )
            variance_prediction *= self.y_std ** 2

            for i in range(len(network_outputs)):
                network_outputs[i] = zero_mean_unit_var_denormalization(
                    network_outputs[i], self.y_mean, self.y_std
                )

        if return_individual_predictions:
            return mean_prediction, variance_prediction, network_outputs[:, :, 0]

        return mean_prediction, variance_prediction

    def predict_single(self, x_test: np.ndarray, sample_index: int):
        """
        Compute the prediction of a single weight sample

        :param x_test: test datapoint
        :param sample_index: specifies the index of the weight sample
        :return: mean and variance of the neural network
        """
        x_test_ = np.asarray(x_test)

        if self.do_normalize_input:
            x_test_, *_ = self.normalize_input(x_test_, self.x_mean, self.x_std)

        def network_predict(x_test_, weights):
            with torch.no_grad():
                self.network_weights = weights
                if self.use_double_precision:
                    return self.model(torch.from_numpy(x_test_).double()).numpy()
                else:
                    return self.model(torch.from_numpy(x_test_).float()).numpy()

        logging.debug("Predicting with %d networks." % len(self.sampled_weights))
        function_value = np.array(network_predict(x_test_, weights=self.sampled_weights[sample_index]))

        if self.do_normalize_output:
            function_value = zero_mean_unit_var_denormalization(
                function_value, self.y_mean, self.y_std
            )
        return function_value

    def f_gradient(self, x_test, weights):
        x_test_ = np.asarray(x_test)

        with torch.no_grad():
            self.network_weights = weights

        if self.use_double_precision:
            x = torch.autograd.Variable(torch.from_numpy(x_test_[None, :]).double(), requires_grad=True)
        else:
            x = torch.autograd.Variable(torch.from_numpy(x_test_[None, :]).float(), requires_grad=True)

        if self.do_normalize_input:
            if self.use_double_precision:
                x_mean = torch.autograd.Variable(torch.from_numpy(self.x_mean).double(), requires_grad=False)
                x_std = torch.autograd.Variable(torch.from_numpy(self.x_std).double(), requires_grad=False)
            else:
                x_mean = torch.autograd.Variable(torch.from_numpy(self.x_mean).float(), requires_grad=False)
                x_std = torch.autograd.Variable(torch.from_numpy(self.x_std).float(), requires_grad=False)

            x_norm = (x - x_mean) / x_std
            m = self.model(x_norm)[0][0]
        else:
            m = self.model(x)[0][0]
        if self.do_normalize_output:

            if self.use_double_precision:
                y_mean = torch.autograd.Variable(torch.from_numpy(np.array([self.y_mean])).double(),
                                                 requires_grad=False)
                y_std = torch.autograd.Variable(torch.from_numpy(np.array([self.y_std])).double(), requires_grad=False)

            else:
                y_mean = torch.autograd.Variable(torch.from_numpy(np.array([self.y_mean])).float(), requires_grad=False)
                y_std = torch.autograd.Variable(torch.from_numpy(np.array([self.y_std])).float(), requires_grad=False)

            m = m * y_std + y_mean

        m.backward()

        g = x.grad.data.numpy()[0, :]
        return g

    def predictive_mean_gradient(self, x_test: np.ndarray):

        # compute the individual gradients for each weight vector
        grads = np.array([self.f_gradient(x_test, weights=weights) for weights in self.sampled_weights])

        # the gradient of the mean is mean of all individual gradients
        g = np.mean(grads, axis=0)

        return g

    def predictive_variance_gradient(self, x_test: np.ndarray):
        m, v, funcs = self.predict(x_test[None, :], return_individual_predictions=True)

        grads = np.array([self.f_gradient(x_test, weights=weights) for weights in self.sampled_weights])

        dmdx = self.predictive_mean_gradient(x_test)

        g = np.mean([2 * (funcs[i] - m) * (grads[i] - dmdx) for i in range(len(self.sampled_weights))], axis=0)

        return g



import torch
import numpy as np
from torch.optim import Optimizer


class PreconditionedSGLD(Optimizer):
    """ Stochastic Gradient Langevin Dynamics Sampler with preconditioning.
        Optimization variable is viewed as a posterior sample under Stochastic
        Gradient Langevin Dynamics with noise rescaled in each dimension
        according to RMSProp.
    """
    def __init__(self,
                 params,
                 lr=np.float64(1e-2),
                 num_train_points=1,
                 precondition_decay_rate=np.float64(0.99),
                 diagonal_bias=np.float64(1e-5)) -> None:
        """ Set up a SGLD Optimizer.

        Parameters
        ----------
        params : iterable
            Parameters serving as optimization variable.
        lr : float, optional
            Base learning rate for this optimizer.
            Must be tuned to the specific function being minimized.
            Default: `1e-2`.
        precondition_decay_rate : float, optional
            Exponential decay rate of the rescaling of the preconditioner (RMSprop).
            Should be smaller than but nearly `1` to approximate sampling from the posterior.
            Default: `0.99`
        diagonal_bias : float, optional
            Term added to the diagonal of the preconditioner to prevent it from
            degenerating.
            Default: `1e-5`.

        """
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))

        defaults = dict(
            lr=lr, precondition_decay_rate=precondition_decay_rate,
            diagonal_bias=diagonal_bias,
            num_train_points=num_train_points
        )
        super().__init__(params, defaults)

    def step(self, closure=None):
        loss = None

        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for parameter in group["params"]:

                if parameter.grad is None:
                    continue

                state = self.state[parameter]
                lr = group["lr"]
                num_train_points = group["num_train_points"]
                precondition_decay_rate = group["precondition_decay_rate"]  # alpha
                diagonal_bias = group["diagonal_bias"]  # lambda
                gradient = parameter.grad.data * num_train_points

                #  state initialization
                if len(state) == 0:
                    state["iteration"] = 0
                    state["momentum"] = torch.ones_like(parameter)

                state["iteration"] += 1

                #  momentum update
                momentum = state["momentum"]
                momentum_t = momentum * precondition_decay_rate + (1.0 - precondition_decay_rate) * (gradient ** 2)
                state["momentum"] = momentum_t  # V(theta_t+1)

                # compute preconditioner
                preconditioner = (1. / (torch.sqrt(momentum_t) + diagonal_bias))  # G(theta_t+1)

                # standard deviation of the injected noise
                sigma = torch.sqrt(torch.from_numpy(np.array(lr, dtype=type(lr)))) * torch.sqrt(preconditioner)

                mean = 0.5 * lr * (preconditioner * gradient)
                delta = (mean + torch.normal(mean=torch.zeros_like(gradient), std=torch.ones_like(gradient)) * sigma)

                parameter.data.add_(-delta)

        return loss



from copy import deepcopy
from functools import partial
import numpy as np
import torch
import torch.nn as nn
from pybnn.bohamiann import Bohamiann
from pybnn.util.layers import AppendLayer


def get_multitask_network(input_dimensionality: int, n_tasks: int) -> torch.nn.Module:

    class Architecture(torch.nn.Module):
        def __init__(self, n_inputs, n_tasks, emb_dim=5, n_hidden=50):
            super(Architecture, self).__init__()
            self.fc1 = torch.nn.Linear(n_inputs - 1 + emb_dim, n_hidden)
            self.fc2 = torch.nn.Linear(n_hidden, n_hidden)
            self.fc3 = torch.nn.Linear(n_hidden, 1)
            self.log_std = AppendLayer(noise=1e-3)
            self.emb = torch.nn.Embedding(n_tasks, emb_dim)
            self.n_tasks = n_tasks

        def forward(self, input):
            x = input[:, :-1]
            t = input[:, -1:]
            t_emb = self.emb(t.long()[:, 0])
            x = torch.cat((x, t_emb), dim=1)
            x = torch.tanh(self.fc1(x))
            x = torch.tanh(self.fc2(x))
            x = self.fc3(x)
            return self.log_std(x)

    return Architecture(n_inputs=input_dimensionality, n_tasks=n_tasks)


class MultiTaskBohamiann(Bohamiann):
    def __init__(self,
                 n_tasks: int,
                 get_network=get_multitask_network,
                 normalize_input: bool = True,
                 normalize_output: bool = True,
                 use_double_precision: bool = True,
                 sampling_method: str = "adaptive_sghmc",
                 metrics=(nn.MSELoss,)
                 ) -> None:
        """ Bayesian Neural Network for regression problems.

        Bayesian Neural Networks use Bayesian methods to estimate the posterior
        distribution of a neural network's weights. This allows to also
        predict uncertainties for test points and thus makes Bayesian Neural
        Networks suitable for Bayesian optimization.
        This module uses stochastic gradient MCMC methods to sample
        from the posterior distribution.

        See [1] for more details.

        [1] J. T. Springenberg, A. Klein, S. Falkner, F. Hutter
            Bayesian Optimization with Robust Bayesian Neural Networks.
            In Advances in Neural Information Processing Systems 29 (2016).

        Parameters
        ----------
        normalize_input: bool, optional
            Specifies if inputs should be normalized to zero mean and unit variance.
        normalize_output: bool, optional
            Specifies whether outputs should be un-normalized.
        """
        self.n_tasks = n_tasks

        func = partial(get_network, n_tasks=n_tasks)
        super(MultiTaskBohamiann, self).__init__(func, normalize_output, normalize_input,
                                                 sampling_method, metrics, use_double_precision)

    def normalize_input(self, x, m=None, s=None):
        if m is None:
            m = np.mean(x, axis=0)
        if s is None:
            s = np.std(x, axis=0)
        x_norm = deepcopy(x)
        x_norm[:, :-1] = (x[:, :-1] - m[:-1]) / s[:-1]

        return x_norm, m, s



import torch
from torch.optim import Optimizer


class ConstantSGD(Optimizer):
    def __init__(self,
                 params,
                 batch_size,
                 num_data_points,
                 precondition_decay_rate=0.95) -> None:
        """

        Parameters
        ----------
        params : iterable
            Parameters serving as optimization variable.

        precondition_decay_rate : float, optional
            Exponential decay rate of the rescaling of the preconditioner (RMSprop).
            Should be smaller than but nearly `1` to approximate sampling from the posterior.
            Default: `0.95`

        """
        self.batch_size = batch_size
        self.num_data_points = num_data_points
        self.precondition_decay_rate = precondition_decay_rate
        defaults = dict(precondition_decay_rate=precondition_decay_rate)
        super().__init__(params, defaults)

    def step(self, closure=None, noise=None):
        loss = None

        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for parameter in group["params"]:

                if parameter.grad is None:
                    continue

                state = self.state[parameter]

                if len(state) == 0:
                    state["iteration"] = 0
                    state["epsilon"] = torch.zeros(1)
                    state["v"] = torch.ones_like(parameter).double()
                gradient = parameter.grad.data
                state["iteration"] += 1
                precondition_decay_rate = group["precondition_decay_rate"]

                v = state["v"]
                v_t = v * precondition_decay_rate + (1.0 - precondition_decay_rate) * (gradient ** 2)
                state["v"] = v_t

                # if noise is None:
                epsilon = (2 * self.batch_size) / (self.num_data_points * torch.sqrt(v_t)) # TODO: variance of std???
                # else:
                #     epsilon = 2 * self.batch_size / (self.num_data_points * noise)
                state["epsilon"] = epsilon

                parameter.data.add_(-epsilon * gradient)

        return loss



from setuptools import setup, find_packages

setup(
    name='pybnn',
    version='0.0.4',
    description='Simple python framework for Bayesian neural networks',
    author='Aaron Klein, Moritz Freidank',
    author_email='kleinaa@cs.uni-freiburg.de',
    url="https://github.com/automl/pybnn",
    license='BSD 3-Clause License',
    classifiers=['Development Status :: 4 - Beta'],
    packages=find_packages(),
    python_requires='>=3',
    install_requires=['torch', 'torchvision', 'numpy', 'emcee', 'scipy'],
    extras_require={},
    keywords=['python', 'Bayesian', 'neural networks'],
)



import torch
import numpy as np
from torch.optim import Optimizer


# Pytorch Port of a previous tensorflow implementation in `tensorflow_probability`:
# https://github.com/tensorflow/probability/blob/master/tensorflow_probability/g3doc/api_docs/python/tfp/optimizer/StochasticGradientLangevinDynamics.md
class pSGLDHD(Optimizer):
    """ Stochastic Gradient Langevin Dynamics Sampler with preconditioning.
        Optimization variable is viewed as a posterior sample under Stochastic
        Gradient Langevin Dynamics with noise rescaled in eaach dimension
        according to RMSProp.
    """
    def __init__(self,
                 params,
                 lr=np.float64(1e-2),
                 hyper_lr=np.float64(1e-2),
                 num_train_points=1,
                 precondition_decay_rate=np.float64(0.99),
                 diagonal_bias=np.float64(1e-5)) -> None:
        """ Set up a SGLD Optimizer.

        Parameters
        ----------
        params : iterable
            Parameters serving as optimization variable.
        lr : float, optional
            Base learning rate for this optimizer.
            Must be tuned to the specific function being minimized.
            Default: `1e-2`.
        precondition_decay_rate : float, optional
            Exponential decay rate of the rescaling of the preconditioner (RMSprop).
            Should be smaller than but nearly `1` to approximate sampling from the posterior.
            Default: `0.95`
        num_pseudo_batches : int, optional
            Effective number of minibatches in the data set.
            Trades off noise and prior with the SGD likelihood term.
            Note: Assumes loss is taken as mean over a minibatch.
            Otherwise, if the sum was taken, divide this number by the batch size.
            Default: `1`.
        num_burn_in_steps : int, optional
            Number of iterations to collect gradient statistics to update the
            preconditioner before starting to draw noisy samples.
            Default: `3000`.
        diagonal_bias : float, optional
            Term added to the diagonal of the preconditioner to prevent it from
            degenerating.
            Default: `1e-5`.

        """
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))

        self.hyper_lr = hyper_lr
        self.lr = lr
        defaults = dict(
            lr=lr, precondition_decay_rate=precondition_decay_rate,
            diagonal_bias=diagonal_bias,
            num_train_points=num_train_points
        )
        super().__init__(params, defaults)

    def step(self, closure=None):
        loss = None

        if closure is not None:
            loss = closure()

        # learning rate update
        for group in self.param_groups:
            for parameter in group["params"]:

                if parameter.grad is None:
                    continue

                state = self.state[parameter]
                #  state initialization
                if len(state) == 0:
                    state["iteration"] = 0
                    state["momentum"] = torch.ones_like(parameter).double()
                    # state["lr"] = torch.ones_like(parameter).double() * self.lr
                    state["dudlr"] = torch.zeros_like(parameter).double()

                gradient = parameter.grad.data

                self.lr -= self.hyper_lr * torch.dot(gradient, state["dudlr"])  # additive rule

        for group in self.param_groups:
            for parameter in group["params"]:

                if parameter.grad is None:
                    continue

                state = self.state[parameter]

                num_train_points = group["num_train_points"]
                precondition_decay_rate = group["precondition_decay_rate"]  # alpha
                diagonal_bias = group["diagonal_bias"]  # lambda
                gradient = parameter.grad.data

                #  momentum update
                momentum = state["momentum"]
                momentum_t = momentum * precondition_decay_rate + (1.0 - precondition_decay_rate) * (gradient ** 2)
                state["momentum"] = momentum_t  # V(theta_t+1)

                # compute preconditioner
                preconditioner = (1. / (torch.sqrt(momentum_t) + diagonal_bias))  # G(theta_t+1)

                # injected noise
                random_sample = torch.normal(mean=torch.zeros_like(gradient), std=torch.ones_like(gradient))
                # random_sample = 0
                # num_train_points = 1
                # parameter update
                sigma = np.sqrt(self.lr) * torch.sqrt(preconditioner)
                mean = 0.5 * self.lr * preconditioner * gradient * num_train_points
                delta = (mean + random_sample * sigma)

                parameter.data.add_(-delta)

                # state["lr"] = lr
                state["dudlr"] = - 0.5 * gradient * preconditioner * num_train_points \
                        + random_sample * torch.sqrt(preconditioner) * 0.5 * 1 / np.sqrt(self.lr)

                state["iteration"] += 1

        return loss



import abc
import numpy as np


class BaseModel(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """
        Abstract base class for all models
        """
        self.X = None
        self.y = None

    @abc.abstractmethod
    def train(self, X, y):
        """
        Trains the model on the provided data.

        Parameters
        ----------
        X: np.ndarray (N, D)
            Input data points. The dimensionality of X is (N, D),
            with N as the number of points and D is the number of input dimensions.
        y: np.ndarray (N,)
            The corresponding target values of the input data points.
        """
        pass

    def update(self, X, y):
        """
        Update the model with the new additional data. Override this function if your
        model allows to do something smarter than simple retraining

        Parameters
        ----------
        X: np.ndarray (N, D)
            Input data points. The dimensionality of X is (N, D),
            with N as the number of points and D is the number of input dimensions.
        y: np.ndarray (N,)
            The corresponding target values of the input data points.
        """
        X = np.append(self.X, X, axis=0)
        y = np.append(self.y, y, axis=0)
        self.train(X, y)

    @abc.abstractmethod
    def predict(self, X_test):
        """
        Predicts for a given set of test data points the mean and variance of its target values

        Parameters
        ----------
        X_test: np.ndarray (N, D)
            N Test data points with input dimensions D

        Returns
        ----------
        mean: ndarray (N,)
            Predictive mean of the test data points
        var: ndarray (N,)
            Predictive variance of the test data points
        """
        pass

    def _check_shapes_train(func):
        def func_wrapper(self, X, y, *args, **kwargs):
            assert X.shape[0] == y.shape[0]
            assert len(X.shape) == 2
            assert len(y.shape) == 1
            return func(self, X, y, *args, **kwargs)
        return func_wrapper

    def _check_shapes_predict(func):
        def func_wrapper(self, X, *args, **kwargs):
            assert len(X.shape) == 2
            return func(self, X, *args, **kwargs)

        return func_wrapper

    def get_json_data(self):
        """
        Json getter function'

        Returns
        ----------
            dictionary
        """
        json_data = {'X': self.X if self.X is None else self.X.tolist(),
                     'y': self.y if self.y is None else self.y.tolist(),
                     'hyperparameters': ""}
        return json_data

    def get_incumbent(self):
        """
        Returns the best observed point and its function value

        Returns
        ----------
        incumbent: ndarray (D,)
            current incumbent
        incumbent_value: ndarray (N,)
            the observed value of the incumbent
        """
        best_idx = np.argmin(self.y)
        return self.X[best_idx], self.y[best_idx]
